/*
 *	kcore - Grab /proc/kcore and dump to a file.
 *
 *	Copyright 1999 (C), Tigran Aivazian.
 *	This is FREE SOFTWARE (under GPL version 2).
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <elf.h>

/* no non-IA32 archs for now */
#define PAGE_SHIFT	12
#define PAGE_SIZE	(1UL << PAGE_SHIFT)
#define PAGE_MASK       (~(PAGE_SIZE-1))
#define PAGE_ALIGN(addr)        (((addr)+PAGE_SIZE-1)&PAGE_MASK)

static int verbose = 1;
static char *kcore_src = "/proc/kcore";
static char *kcore_dst = "kcore.img";

static void usage(void)
{
	fprintf(stderr, "kcore 0.1, by Tigran Aivazian <tigran@sco.com>\n");
	fprintf(stderr, "usage: kcore [-h] [-q] [-V] [-k kcore_src] [-f kcore_dst]\n");
	exit(1);
}

static void parse_args(int argc, char *argv[])
{
	int c;

	while (EOF != (c = getopt(argc, argv, "hvk:f:V"))) {
		switch (c) {
			case 'h':
				usage();
				break;

			case 'k':
				kcore_src = strdup(optarg);
				break;

			case 'f':
				kcore_dst = strdup(optarg);
				break;

			case 'q':
				verbose = 0;
				break;

			case 'V':
				fprintf(stderr, "Version 0.1\n");
				exit(1);

			default:
				usage();
				break;
		}
	}
}

static int open_kcore(void)
{
	int fd;

	fd = open(kcore_src, O_RDONLY);
	if (fd == -1) {
		fprintf(stderr, "kcore: open(%s), errno=%d (%s)\n",
			kcore_src, errno, strerror(errno));
		exit(1);
	}
	return fd;
}

static int open_dump(void)
{
	int fd;

	(void)umask(0);
	fd = open(kcore_dst, O_RDWR|O_CREAT|O_TRUNC, 0600);
	if (fd == -1) {
		fprintf(stderr, "kcore: open(%s), errno=%d (%s)\n",
			kcore_dst, errno, strerror(errno));
		exit(1);
	}
	return fd;
}

static ssize_t Read(int fd, void *buf, size_t count)
{
	if (read(fd, buf, count) != count) {
		fprintf(stderr, "kcore: read(%d,%08lx,%d), errno=%d (%s)\n",
			fd, (unsigned long)buf, count, errno, strerror(errno));
		exit(1);
	}
	return count;
}

static ssize_t Write(int fd, void *buf, size_t count)
{
	if (write(fd, buf, count) != count) {
		fprintf(stderr, "kcore: write(%d,%08lx,%d), errno=%d (%s)\n",
			fd, (unsigned long)buf, count, errno, strerror(errno));
		exit(1);
	}
	return count;
}

off_t Lseek(int fd, off_t offset)
{
	if (lseek(fd, offset, SEEK_SET) == -1) {
		fprintf(stderr, "kcore: lseek(%d, %ld), errno=%d (%s)\n",
			fd, offset, errno, strerror(errno));
		exit(1);
	}
	return 0;
}

static void *Malloc(size_t size)
{
	char *buf = malloc(size);
	if (!buf) {
		fprintf(stderr, "kcore: malloc(%d), errno=%d (%s)\n",
			size, errno, strerror(errno));
		exit(1);
	}
	return buf;
}

static void dump_kcore(int kfd, int fd)
{
	Elf32_Ehdr elf;
	Elf32_Phdr phdr;
	unsigned long dataoff;
	char * buf;
	int i;

	Read(kfd, &elf, sizeof(elf));
	if (memcmp(&elf, ELFMAG, SELFMAG) || elf.e_type != ET_CORE) {
		fprintf(stderr, "kcore: No ELF core magic in %s\n", kcore_src);
		exit(1);
	}
	dataoff = PAGE_ALIGN(elf.e_phoff + elf.e_phentsize * elf.e_phnum);
	buf = Malloc(dataoff);
	Lseek(kfd, 0);
	Read(kfd, buf, dataoff);
	Write(fd, buf, dataoff);
	free(buf);
	if (verbose)
		fprintf(stderr, "kcore: dumped header: %08lx - %08lx (%ld bytes)\n", 
			0L, dataoff, dataoff);
	for (i=0; i<elf.e_phnum; i++) {
		Lseek(kfd, elf.e_phoff + i*elf.e_phentsize);
		Read(kfd, &phdr, sizeof(phdr));
		Lseek(kfd, phdr.p_offset);
		buf = Malloc(phdr.p_filesz);
		Read(kfd, buf, phdr.p_filesz);
		Lseek(fd, phdr.p_offset);
		Write(fd, buf, phdr.p_filesz);
		free(buf);
		if (verbose)
			fprintf(stderr, "kcore: dumped %s segment: %08lx - %08lx (%ld bytes)\n",
				phdr.p_type == PT_NOTE ? "PT_NOTE" : "PT_LOAD", 
				(unsigned long)phdr.p_offset, 
				(unsigned long)(phdr.p_offset + phdr.p_filesz),
				(unsigned long)phdr.p_filesz);
	}
}

int main(int argc, char *argv[])
{
	int kfd, fd;

	parse_args(argc, argv);	
	kfd = open_kcore();
	fd = open_dump();
	dump_kcore(kfd, fd);
	(void)close(kfd);
	(void)close(fd);
	return 0;
}
